import collections

import numpy as np


FLOAT_32_LE = np.dtype("<f4")
RSC_KEY_TYPES = [
    ("width", int),
    ("file_length", int),
    ("x_first", float),
    ("y_first", float),
    ("x_step", float),
    ("y_step", float),
    ("x_unit", str),
    ("y_unit", str),
    ("z_offset", int),
    ("z_scale", int),
    ("projection", str),
]


def load(
    filename,
    rsc_file=None,
    rows=None,
    cols=None,
    band=1,
    **kwargs,
):
    """Load a file, either using numpy or rasterio"""
    if rsc_file:
        rsc_data = load_rsc(rsc_file)
        return load_stacked_img(filename, rsc_data=rsc_data)
    else:
        try:
            import rasterio as rio
        except ImportError:
            raise ValueError("Need to `conda install rasterio` to load gdal-readable")

        with rio.open(filename) as src:
            return src.read(band)


def load_rsc(filename, lower=False, **kwargs):
    """Loads and parses the .rsc file

    Args:
        filename (str) path to either the or .rsc file.
            Function will add .rsc to path if passed .dem file
        lower (bool): make keys of the dict lowercase

    Returns:
        dict: .rsc file parsed out

    example file:
    WIDTH         10801
    FILE_LENGTH   7201
    X_FIRST       -157.0
    Y_FIRST       21.0
    X_STEP        0.000277777777
    Y_STEP        -0.000277777777
    X_UNIT        degrees
    Y_UNIT        degrees
    Z_OFFSET      0
    Z_SCALE       1
    PROJECTION    LL
    """

    # Use OrderedDict so that upsample_dem_rsc creates with same ordering as old
    output_data = collections.OrderedDict()
    # Second part in tuple is used to cast string to correct type

    with open(filename, "r") as f:
        for line in f.readlines():
            for field, num_type in RSC_KEY_TYPES:
                if line.startswith(field.upper()):
                    output_data[field] = num_type(line.split()[1])

    if lower:
        output_data = {k.lower(): d for k, d in output_data.items()}
    return output_data


def load_stacked_img(
    filename,
    arr=None,
    rows=None,
    cols=None,
    rsc_data=None,
    return_amp=False,
    dtype=FLOAT_32_LE,
    **kwargs,
):
    """Helper function to load .unw and .cor files from snaphu output

    Format is two stacked matrices:
        [[first], [second]] where the first "cols" number of floats
        are the first matrix, next "cols" are second, etc.
    Also called BIL, Band Interleaved by Line
    See http://webhelp.esri.com/arcgisdesktop/9.3/index.cfm?topicname=BIL,_BIP,_and_BSQ_raster_files
    for explantion

    For .unw height files, the first is amplitude, second is phase (unwrapped)
    For .cc correlation files, first is amp, second is correlation (0 to 1)

    Args:
        filename (str): path to the file to open
        rows (int): manually pass number of rows (overrides rsc/ann data)
        cols (int): manually pass number of cols (overrides rsc/ann data)
        rsc_data (dict): output from load_rsc, gives width of file
        return_amp (bool): flag to request the amplitude data to be returned

    Returns:
        ndarray: dtype=float32, the second matrix (height, correlation, ...) parsed
        if return_amp == True, returns two ndarrays stacked along axis=0
    """
    if rows is None or cols is None:
        rows, cols = rsc_data["file_length"], rsc_data["width"]

    data = np.fromfile(filename, FLOAT_32_LE)

    first = data.reshape((rows, 2 * cols))[:, :cols]
    second = data.reshape((rows, 2 * cols))[:, cols:]
    if return_amp:
        return np.stack((first, second), axis=0)
    else:
        return second



import argparse
from . import core


description = """
'--level pixel' means individual pixels for each SAR date are labeled (good for larger scenes).
'--level scene' means whole SAR images are labeled (good for smaller scenes).
For scene level labeling, the variance of each average interferogram is used.
"""


def get_cli_args():
    p = argparse.ArgumentParser(description=description)
    p.add_argument(
        "--level",
        default="pixel",
        choices=["pixel", "scene"],
        help=("Level at which to label outliers. (default=%(default)s).\n"),
    )
    p.add_argument(
        "--ext",
        default=".unw",
        help="filename extension of unwrapped igrams to average (default=%(default)s)",
    )
    p.add_argument(
        "--search-path",
        "-p",
        default=".",
        help="location of igram files. (default=%(default)s)",
    )
    p.add_argument(
        "--outfile",
        "-o",
        default="labels.nc",
        help="Location to save final labels (default=%(default)s)",
    )
    p.add_argument(
        "--avg-file",
        default="average_slcs.nc",
        help="Location to save stack of averaged igrams (default=%(default)s)",
    )
    p.add_argument(
        "--deramp",
        action="store_true",
        default=True,
        help="remove a linear ramp from phase after averaging (default=%(default)s)",
    )
    p.add_argument(
        "--nsigma",
        "-n",
        help=(
            "Number of sigma_mad deviations away from median to label as outlier"
            " (default=%(default)s)"
        ),
        default=5,
    )
    p.add_argument(
        "--rsc-file", help="If using ROI_PAC .rsc files, location of .rsc file"
    )
    p.add_argument(
        "--overwrite",
        action="store_true",
        help="Overwrite existing averaged files (default=%(default)s)",
    )
    p.add_argument(
        "--normalize-time",
        action="store_true",
        help="Divide igram phase by temporal baseline (default=%(default)s)",
    )
    return p.parse_args()


def average_and_label():
    args = get_cli_args()
    core.create_averages(**vars(args))
    core.label_outliers(
        fname=args.avg_file,
        outfile=args.outfile,
        nsigma=args.nsigma,
        level=args.level,
    )



"""
This module exports a Log class that wraps the logging python package

Uses the standard python logging utilities, just provides
nice formatting out of the box.

Usage:

    from apertools.log import get_log
    logger = get_log()

    logger.info("Something happened")
    logger.warning("Something concerning happened")
    logger.error("Something bad happened")
    logger.critical("Something just awful happened")
    logger.debug("Extra printing we often don't need to see.")
    # Custom output for this module:
    logger.success("Something great happened: highlight this success")
"""
from functools import wraps
import logging
import time

try:
    from colorlog import ColoredFormatter

    COLORS = True
except ImportError:
    from logging import Formatter

    COLORS = False


def get_log(debug=False, name=__file__, verbose=False):
    logger = logging.getLogger(name)
    return format_log(logger, debug=debug, verbose=verbose)


def format_log(logger, debug=False, verbose=False):
    log_level = logging.DEBUG if debug else logging.INFO
    log_colors = {
        "DEBUG": "blue",
        "INFO": "black",
        "WARNING": "red",
        "ERROR": "red",
        "CRITICAL": "black,bg_red",
    }

    if COLORS:
        format_ = "[%(asctime)s] [%(log_color)s%(levelname)s %(filename)s%(reset)s] %(message)s%(reset)s"
        formatter = ColoredFormatter(
            format_, datefmt="%m/%d %H:%M:%S", log_colors=log_colors
        )
    else:
        format_ = "[%(asctime)s] [%(levelname)s %(filename)s] %(message)s"
        formatter = Formatter(format_, datefmt="%m/%d %H:%M:%S")

    handler = logging.StreamHandler()
    handler.setFormatter(formatter)

    if not logger.handlers:
        logger.addHandler(handler)
        logger.setLevel(log_level)

        if verbose:
            logger.info("Logger initialized: %s" % (logger.name,))

    if debug or verbose:
        logger.setLevel(logging.DEBUG)

    return logger


logger = get_log()


def log_runtime(f):
    """
    Logs how long a decorated function takes to run
    Args:
        f (function): The function to wrap
    Retruns:
        function: The wrapped function
    """
    # pass function's docstring through
    @wraps(f)
    def wrapper(*args, **kwargs):
        t1 = time.time()
        result = f(*args, **kwargs)
        t2 = time.time()
        elapsed_time = t2 - t1

        time_string = "Total elapsed time for {} : {} minutes ({} seconds)".format(
            f.__name__,
            "{0:.2f}".format(elapsed_time / 60.0),
            "{0:.2f}".format(elapsed_time),
        )
        logger.info(time_string)
        return result

    return wrapper



#!/usr/bin/env python
"""
Create labels of outlier interferogram pixels.

Uses the averaged unwrapped igrams per date.
"""
import os
import numpy as np

from . import utils
from . import sario
from .logger import get_log, log_runtime
from .deramp import remove_ramp

log = get_log()


def label_outliers(
    fname=None,
    stack=None,
    outfile="labels.nc",
    nsigma=5,
    level="pixel",
    min_spread=0.5,
):
    if stack is None:
        import xarray as xr

        stack = xr.open_dataarray(fname)
    log.info("Computing {} sigma outlier labels at {} level.".format(nsigma, level))

    if level == "pixel":
        # Use all pixel absolute values here, shape: (ndates, rows, cols)
        stack_data = np.abs(stack)
        labels, threshold = label(stack_data, nsigma=nsigma, min_spread=min_spread)
    elif level == "scene":
        # Use just scene-level variance, shape: (ndates, 1, 1)
        stack_data = np.var(stack, axis=(1, 2), keepdims=True)
        labels, threshold = label(stack_data, nsigma=nsigma, min_spread=min_spread)
        # Add squeeze for the scene-level case, dont need lat/lon dims
        labels = labels.squeeze()
        stack_data = stack_data.squeeze()
        threshold = threshold.squeeze()
    else:
        raise ValueError("`level` must be 'pixel' or 'scene'")

    # Rename the xarray dataarrays
    labels = labels.rename("labels")
    stack_data = stack_data.rename("data")
    threshold = threshold.rename("threshold")
    if outfile:
        log.info("Saving outlier labels to {}:/labels".format(outfile))
        labels.to_netcdf(outfile)
        log.info("Saving data to {}:/data".format(outfile))
        stack_data.to_netcdf(outfile, mode="a")
        log.info("Saving threshold to {}:/threshold".format(outfile))
        threshold.to_netcdf(outfile, mode="a")
    return labels, threshold


def label(
    data,
    nsigma=5,
    min_spread=0.5,
):
    med = data.median(axis=0)
    spread = np.maximum(min_spread, nsigma * mad(data, axis=0))
    threshold = med + spread
    return (data > threshold), threshold


def mad(stack, axis=0, scale=1.4826):
    """Median absolute deviation,

    default is scaled such that +/-MAD covers 50% (between 1/4 and 3/4)
    of the standard normal cumulative distribution
    """
    stack_abs = np.abs(stack)
    med = np.nanmedian(stack_abs, axis=axis)
    return scale * np.nanmedian(np.abs(stack_abs - med), axis=axis)


@log_runtime
def create_averages(
    search_path=".",
    ext=".unw",
    rsc_file=None,
    deramp=True,
    avg_file="average_slcs.nc",
    overwrite=False,
    normalize_time=False,
    band=2,
    ds_name="igrams",
    **kwargs,
):
    """Create a NetCDF stack of "average interferograms" for each date

    Args:
        search_path (str):
            directory to find igrams
        ext (str):
            extension name of unwrapped interferograms (default = .unw)
        rsc_file (str):
            filename of .rsc resource file, if loading binary files like snaphu outputs
        deramp (bool):
            remove a linear ramp from unwrapped igrams when averaging
        avg_file (str):
            name of output file to save stack
        overwrite (bool):
            clobber current output file, if exists
        normalize_time (bool):
            Divide igram phase by temporal baseline (default = false)
            true: units = [rad / day], false: units = [rad]
        band (int):
            if using rasterio to load igrams, which image band to load
        ds_name (str):
            Name of the data variable used in the netcdf stack
    """
    import netCDF4 as nc

    if os.path.exists(avg_file) and not overwrite:
        log.info("{} exists, not overwriting.".format(avg_file))
        return avg_file

    log.info("Searching for igrams in {} with extention {}".format(search_path, ext))
    ifg_date_list = utils.find_igrams(directory=search_path, ext=ext)
    unw_file_list = utils.find_igrams(directory=search_path, ext=ext, parse=False)
    sar_date_list = utils.dates_from_igrams(ifg_date_list)

    nigrams, ndates = len(ifg_date_list), len(sar_date_list)
    log.info("Found {} igrams, {} unique SAR dates".format(nigrams, ndates))

    utils.create_empty_nc_stack(
        avg_file,
        date_list=sar_date_list,
        rsc_file=rsc_file,
        gdal_file=unw_file_list[0],
        stack_data_name=ds_name,
        overwrite=overwrite,
    )

    f = nc.Dataset(avg_file, mode="r+")
    ds = f["igrams"]
    _, rows, cols = ds.shape

    # TODO: support masks or not?
    # Get masks for deramping
    # mask_igram_date_list = utils.load_intlist_from_h5(mask_fname)
    out_mask = np.zeros((rows, cols)).astype(bool)

    for (idx, gdate) in enumerate(sar_date_list):
        cur_unws = [
            (f, date_pair)
            for (date_pair, f) in zip(ifg_date_list, unw_file_list)
            if gdate in date_pair
        ]
        log.info(
            "Averaging {} igrams for {} ({} out of {})".format(
                len(cur_unws), gdate, idx + 1, len(sar_date_list)
            )
        )

        # reset the matrix to all zeros
        out = 0
        for unwf, date_pair in cur_unws:
            img = sario.load(unwf, rsc_file=rsc_file, band=band)
            if normalize_time:
                img /= (date_pair[1] - date_pair[0]).days
            out += img

            # mask_idx = mask_igram_date_list.index(date_pair)
            # out_mask |= mask_stack[mask_idx]

        out /= len(cur_unws)

        if deramp:
            out = remove_ramp(out, deramp_order=1, mask=out_mask)
        else:
            out[out_mask] = np.nan

        # Write the single layer out
        ds[idx, :, :] = out

    # Close to save it
    f.close()
    return avg_file



import os
import datetime
import time
from collections.abc import Iterable
from glob import glob
import numpy as np
import netCDF4 as nc
import itertools

from .logger import get_log
from . import sario

log = get_log()
DATE_FMT = "%Y%m%d"


def find_igrams(directory=".", ext=".int", parse=True, filename=None):
    """Reads the list of igrams to return dates of images as a tuple

    Args:
        directory (str): path to the igram directory
        ext (str): file extension when searching a directory
        parse (bool): output as parsed datetime tuples. False returns the filenames
        filename (str): name of a file with SAR filenames separated by newlines

    Returns:
        tuple(date, date) of (early, late) dates for all igrams (if parse=True)
            if parse=False: returns list[str], filenames of the igrams

    """
    if filename is not None:
        with open(filename) as f:
            igram_file_list = [
                line
                for line in f.read().splitlines()
                if not line.strip().startswith("#")
            ]
    else:
        igram_file_list = sorted(glob(os.path.join(directory, "*" + ext)))

    if parse:
        igram_fnames = [os.path.split(f)[1] for f in igram_file_list]
        date_pairs = [intname.strip(ext).split("_")[:2] for intname in igram_fnames]
        return parse_intlist_strings(date_pairs, ext=ext)
    else:
        return igram_file_list


def parse_intlist_strings(date_pairs, ext=".int"):
    # If we passed filename YYYYmmdd_YYYYmmdd.int
    if not date_pairs:
        return []
    if isinstance(date_pairs, str):
        date_pairs = [date_pairs.split(".")[0].split("_")[:2]]
    elif isinstance(date_pairs, Iterable) and isinstance(date_pairs[0], str):
        date_pairs = [f.split(".")[0].split("_")[:2] for f in date_pairs]

    return [(_parse(early), _parse(late)) for early, late in date_pairs]


def dates_from_igrams(igram_list):
    """Takes a list of [(reference, secondary),...] igram date pairs
    and returns the list of unique dates of SAR images used to form them
    """
    return sorted(list(set(itertools.chain(*igram_list))))


def _parse(datestr):
    return datetime.datetime.strptime(datestr, DATE_FMT).date()


def get_latlon_arrs(h5_filename=None, rsc_file=None, gdal_file=None):
    if rsc_file is not None:
        lon_arr, lat_arr = grid(**sario.load_rsc(rsc_file), sparse=True)
    elif gdal_file is not None:
        lon_arr, lat_arr = grid(fname=gdal_file)

    lon_arr, lat_arr = lon_arr.reshape(-1), lat_arr.reshape(-1)
    return lon_arr, lat_arr


def grid(
    rows=None,
    cols=None,
    y_step=None,
    x_step=None,
    y_first=None,
    x_first=None,
    width=None,
    file_length=None,
    sparse=True,
    fname=None,
    **kwargs,
):
    """Takes sizes and spacing info, creates a grid of values

    Args:
        rows (int): number of rows
        cols (int): number of cols
        y_step (float): spacing between rows
        x_step (float): spacing between cols
        y_first (float): starting location of first row at top
        x_first (float): starting location of first col on left
        sparse (bool): Optional (default False). Passed through to
            np.meshgrid to optionally conserve memory

    Returns:
        tuple[ndarray, ndarray]: the XX, YY grids of longitudes and lats

    Examples:
    >>> test_grid_data = {'cols': 2, 'rows': 3, 'x_first': -155.0, 'x_step': 0.01,\
'y_first': 19.5, 'y_step': -0.2}
    >>> lons, lats = grid(**test_grid_data)
    >>> np.set_printoptions(legacy="1.13")
    >>> print(lons)
    [[-155.   -154.99]
     [-155.   -154.99]
     [-155.   -154.99]]
    >>> print(lats)
    [[ 19.5  19.5]
     [ 19.3  19.3]
     [ 19.1  19.1]]
    """
    if fname is None:
        rows = rows or file_length
        cols = cols or width
        x = np.linspace(x_first, x_first + (cols - 1) * x_step, cols).reshape((1, cols))
        y = np.linspace(y_first, y_first + (rows - 1) * y_step, rows).reshape((rows, 1))
    else:
        try:
            import rasterio as rio
        except ImportError:
            raise ValueError(
                "Need to `conda install rasterio` to pass gdal-readable files to `grid`"
            )

        with rio.open(fname) as src:
            rows, cols = src.shape
            max_len = max(rows, cols)
            lon_list, lat_list = src.xy(np.arange(max_len), np.arange(max_len))
        x = np.arange(lon_list[:cols])
        y = np.arange(lat_list[:rows])

    return np.meshgrid(x, y, sparse=sparse)


def create_empty_nc_stack(
    outname,
    date_list=None,
    rsc_file=None,
    gdal_file=None,
    dtype="float32",
    stack_dim_name="date",
    stack_data_name="igrams",
    lat_units="degrees north",
    lon_units="degrees east",
    overwrite=False,
):
    """Creates skeleton of .nc stack without writing stack data

    Args:
        outname (str): name of .nc output file to save
        date_list (list[datetime.date]): if layers of stack correspond to dates of SAR images
        rsc_file (str): .rsc (resource) file containing the desired output lat/lon grid data
        gdal_file (str): instead of .rsc, and example GDAL-readable file in desired coordinates
        dtype: default="float32", the numpy datatype of the stack data
        stack_dim_name (str): default = "date". Name of the 3rd dimension of the stack
            (Dimensions are (stack_dim_name, lat, lon) )
        stack_data_name (str): default="stack", name of the data variable in the file
        lat_units (str): default = "degrees north",
        lon_units (str): default = "degrees east",
        overwrite (bool): default = False, will overwrite file if true
    """
    if not outname.endswith(".nc"):
        raise ValueError("{} must be an .nc filename".format(outname))

    # TODO: allow for radar coordinates and just "x, y" generic?
    lon_arr, lat_arr = get_latlon_arrs(
        rsc_file=rsc_file,
        gdal_file=gdal_file,
    )

    rows, cols = len(lat_arr), len(lon_arr)
    depth = len(date_list)

    if date_list is None:
        raise ValueError("Need 'date_list' if 3rd dimension is 'date'")
    stack_dim_arr = to_datetimes(date_list)

    log.info("Making dimensions and variables")
    with nc.Dataset(outname, "w", clobber=overwrite) as f:
        f.history = "Created " + time.ctime(time.time())

        f.createDimension("lat", rows)
        f.createDimension("lon", cols)
        # Could make this unlimited to add to it later?
        latitudes = f.createVariable("lat", "f4", ("lat",), zlib=True)
        longitudes = f.createVariable("lon", "f4", ("lon",), zlib=True)
        latitudes.units = "degrees north"
        longitudes.units = "degrees east"

        f.createDimension(stack_dim_name, depth)
        stack_dim_variable = f.createVariable(
            stack_dim_name, "f4", (stack_dim_name,), zlib=True
        )
        stack_dim_variable.units = "days since {}".format(date_list[0])

        # Write data
        latitudes[:] = lat_arr
        longitudes[:] = lon_arr
        d2n = nc.date2num(stack_dim_arr, units=stack_dim_variable.units)
        stack_dim_variable[:] = d2n

        # Finally, the actual stack
        # stackvar = rootgrp.createVariable("stack/1", "f4", ("date", "lat", "lon"))
        log.info("Writing dummy data for %s", stack_data_name)
        dt = np.dtype(dtype)
        fill_value = 0

        f.createVariable(
            stack_data_name,
            dt,
            (stack_dim_name, "lat", "lon"),
            fill_value=fill_value,
            zlib=True,
        )


def to_datetimes(date_list):
    return [datetime.datetime(*d.timetuple()[:6]) for d in date_list]


# def ignore_sar_dates(
#     sar_date_list, int_date_list, ignore_file="sarlist_ignore.txt", parse=True
# ):
#     """Read extra file to ignore certain dates of interferograms"""
#     ignore_sars = set(find_sars(filename=ignore_file, parse=parse))
#     log.info("Ignoring the following .sar dates:")
#     log.info(sorted(ignore_sars))
#     valid_sars = [g for g in sar_date_list if g not in ignore_sars]
#     valid_igrams = [
#         i for i in int_date_list if i[0] not in ignore_sars and i[1] not in ignore_sars
#     ]
#     return valid_sars, valid_igrams
